"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a sum of terms that all contain the correct number
of test function terms (which is equal to the form rank).
"""
from __future__ import absolute_import

from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from ufl.algorithms import MultiFunction


@ufl_transformation(name="accterms2", extraction_lambda=lambda l: [i[0] for i in l])
def split_into_accumulation_terms(expr):
    mod_args = extract_modified_arguments(expr)

    accumulation_terms = []
    for arg in mod_args:
        from dune.perftool.ufl.transformations.replace import replace_expression
        from ufl.classes import Zero, IntValue
        # Define a replacement map that maps the given arg to 1 and the rest to 0
        rmap = {ma: Zero() for ma in mod_args}
        rmap[arg] = IntValue(1)

        # Do the replacement on the expression
        accum_expr = replace_expression(expr, rmap)

        # Store the foudn accumulation expression
        accumulation_terms.append((accum_expr, arg))

    return accumulation_terms


# class SplitIntoAccumulationTerms(MultiFunction):
#     """ return a list of tuples of expressions and modified arguments! """
#
#     call = MultiFunction.__call__
#
#     def __call__(self, expr):
#         from dune.perftool.ufl.rank import ufl_rank
#         self.rank = ufl_rank(expr)
#
#         from dune.perftool.ufl.modified_terminals import ModifiedArgumentExtractor
#         self.mae = ModifiedArgumentExtractor()
#
#         from dune.perftool.ufl.transformations.argument_elimination import EliminateArguments
#         self.ea = EliminateArguments()
#
#         # Collect the found terms
#         self.terms = {}
#
#         # Now, fill the terms dict by applying this multifunction!
#         self.call(expr)
#
#         from dune.perftool.ufl.flatoperators import construct_binary_operator
#         from ufl.classes import Sum
#         return [(construct_binary_operator(t, Sum), a) for a, t in self.terms.items()]
#
#     def expr(self, o):
#         # This needs to be a valid accumulation term!
#         assert all(len(self.mae(o, i)) == 1 for i in range(self.rank))
#
#         expr, arg = self.ea(o)
#         if arg not in self.terms:
#             self.terms[arg] = []
#         self.terms[arg].append(expr)
#
#     def sum(self, o):
#         # Check whether this sums contains too many accumulation terms.
#         if not all(len(self.mae(o, i)) == 1 for i in range(self.rank)):
#             # This sum is part of a top level sum that separates accumulation terms!
#             for op in o.ufl_operands:
#                 self.call(op)
#         else:
#             # This is a normal sum, we might treat it as any other expression
#             self.expr(o)
#
#     def index_sum(self, o):
#         # Check whether this sums contains too many accumulation terms.
#         if not all(len(self.mae(o, i)) == 1 for i in range(self.rank)):
#             # This sum is part of a top level sum that separates accumulation terms!
#             self.call(o.ufl_operands[0])
#         else:
#             # This is a normal sum, we might treat it as any other expression
#             # TODO we need to eliminate topsum indexsum regardless of the thing being valid
#             self.expr(o.ufl_operands[0])
#             # old code
#             #self.expr(o)
#
#
# from dune.perftool.ufl.transformations import ufl_transformation
# @ufl_transformation(name="accterms", extraction_lambda=lambda l:[i[0] for i in l])
# def split_into_accumulation_terms2(expr):
#     return SplitIntoAccumulationTerms()(expr)
